/* InsertFieldUnit.groovy
October 2008 by John Green
 */
import org.prorefactor.treeparser.ParseUnit
import org.prorefactor.core.JPNode
import org.prorefactor.core.TreeUtils
import org.prorefactor.core.IConstants
import org.prorefactor.core.TokenTypes as TT
import org.prorefactor.nodetypes.RecordNameNode
import org.prorefactor.refactor.FileTarget
import org.prorefactor.refactor.FileStuff
import jdbm.btree.BTree


class InsertFieldUnit {

	final boolean CREATE_DISABLED = false
	final boolean RECORDNAME_DISABLED = false

	InsertFieldUnit(ParseUnit parseUnit,
				ScriptSupport support, 
				BTree targetTable,
				HashSet<String> doneFileNames
				) {
		this.parseUnit = parseUnit
		this.support = support
		this.targetTable = targetTable
		this.doneFileNames = doneFileNames
		nodeArray = TreeUtils.nodeArray(parseUnit.topNode)
		this.logger = support.logger
	}

	JPNode [] nodeArray
	ParseUnit parseUnit
	ScriptSupport support
	HashSet<String> doneFileNames = new HashSet<String>()

	int currSourceFileNumChanges = 0
	int totalChangedFiles = 0
	int totalChanges = 0

	def currSourceLines = new ArrayList<String>()

	BTree targetTable
	File currSourceFile = null
	Logger logger
	String relativePath

	final String LINESEP = System.getProperty('line.separator')


	void run() {
		// Use a map that is sorted by FileTarget, so that multiple
		// entries from any one file are all grouped together.
		def fileTargets = new TreeMap()
		String canonName = null
		for (JPNode node in nodeArray) {
			boolean needsConvert
			switch (node.type) {
				case TT.CREATE:
					needsConvert = testCreateNode(node)
					break
				case TT.RECORD_NAME:
					needsConvert = testRecordNameNode((RecordNameNode)node)
					break
				case TT.INSERT:
					needsConvert = testInsertNode(node)
					break
			}
			if (needsConvert)
				canonName = new File(node.filename).canonicalPath
			if (needsConvert && ! doneFileNames.contains(canonName)) {
				def fileTarget = new FileTarget(
						canonName,
						node.line - 1,
						node.column - 1 )
				// Work with FileTarget objects, to avoid duplicate conversion from
				// code in an include file referenced more than once.
				if (! fileTargets.containsKey(fileTarget)) {
					fileTargets.put(fileTarget, node)
				}
			}
		}
		fileTargets.each{ FileTarget target, JPNode node ->
			doneFileNames.add(target.filename)
			if (targetTable.find(target) != null)
				return // break out of this closure instance
			loadIfNotCurrent(node.filename)
			Conversion conversion = null
			switch (node.type) {
				case TT.CREATE:
					conversion = convertCreate(target, node)
					break
				case TT.RECORD_NAME:
					conversion = convertRecordName(target, (RecordNameNode)node)
					break
				case TT.INSERT:
					conversion = new Conversion()
					conversion.reportText = "INSERT " + node.firstChild().text
					break
			}
			targetTable.insert(target, conversion, false)
		}
		if (currSourceFileNumChanges > 0)
			writeCurrent()
	}


	/** Append text to a line.
	 * Input line number must be index-from-zero.
	 */
	void append(int line, String s) {
		def text = currSourceLines[line]
		if (text)
			currSourceLines[line] = text + s
		else
			currSourceLines[line] = s
	}
	
	
	/** Append an indented new line after an existing line. */
	void appendIndentedNewLineToLine(line, String indent, String insertText) {
		String oldText = currSourceLines[line]
		def newText = new StringBuilder()
		newText
			.append(oldText)
			.append(LINESEP)
			.append(indent)
			.append(insertText)
			;
		currSourceLines[line] = newText.toString()
	}
	
	
	/** Append an indented new line after a node's text.
	 * For any text that was on the same line after the node in the original text,
	 * that gets pushed down to a new indented line too.
	 */
	void appendIndentedNewLineToNode(JPNode node, String indent, String insertText) {
		int line = node.line - 1
		String oldText = currSourceLines[line]
		int column = node.column - 1 + node.text.length()
		String textAfter = oldText.substring(column).trim()
		def newText = new StringBuilder()
		newText
			.append(oldText.substring(0, column))
			.append(LINESEP)
			.append(indent)
			.append(insertText)
			;
		if (textAfter) {
			newText
				.append(LINESEP)
				.append(indent)
				.append(textAfter)
				;
						
		}
		currSourceLines[line] = newText.toString()
	}
		
	
	/** Append text to a node's text. */
	void appendToNode(JPNode node, String text) {
		int line = node.line - 1
		String oldLine = currSourceLines[line]
		int column = node.column - 1 + node.text.length()
		currSourceLines[line] = 
			oldLine.substring(0, column) +
			text +
			oldLine.substring(column)
	}


	Conversion convertCreate(FileTarget fileTarget, JPNode createNode) {
		def ret = new Conversion()
		def recordNameNode = (RecordNameNode) createNode.firstChild()
		ret.reportText = 'CREATE ' + recordNameNode.text
		
		JPNode modifyNode = createNode.nextSibling()
		// If there is no ASSIGN statement, or if there is an assignment without ASSIGN keyword/text,
		// then we're going to work from the period at the end of the create statement.
		if (! modifyNode || modifyNode.type != TT.ASSIGN ||  (! modifyNode.text)) {
			modifyNode = createNode.lastDescendant()
			if (modifyNode.type != TT.PERIOD) {
				logNoConvert(createNode, 'Did not find period (.) at end of CREATE statement.')
				ret.preprocessorProblems = true
				return ret
			}
		}
		
		ret.preprocessorProblems = ! matchList([createNode, recordNameNode, modifyNode])
		if (ret.preprocessorProblems || CREATE_DISABLED)
			return ret
			
		///// FROM HERE DOWN MUST BE ATOMIC (ALL OR NOTHING) /////
		currSourceFileNumChanges++

		String insertText = recordNameNode.text + ".entity = roadrunner.System:Entity"

		if (modifyNode.type == TT.ASSIGN) {
			JPNode assignPeriod = modifyNode.lastDescendant()
			if (	! assignPeriod
				||	assignPeriod.type != TT.PERIOD
				||	assignPeriod.fileIndex != modifyNode.fileIndex
				) {
				// We just insert our text at the end of the ASSIGN node's text.
				String newText = " " + insertText + " "
				appendToNode(modifyNode, newText)
			} else {
				String indent = getIndent(modifyNode.line) // is line after, count by 0
				appendIndentedNewLineToNode(modifyNode, indent, insertText)
			}
		} else {
			// Insert a new ASSIGN statement
			String indent = getIndent(createNode.line - 1) // count by zero, line of CREATE
			appendIndentedNewLineToLine(modifyNode.line - 1, indent,
					"ASSIGN " + insertText + "."
					)
		}

		return ret
	} // convertCreate


	Conversion convertRecordName(FileTarget fileTarget, RecordNameNode recordNameNode) {
		def ret = new Conversion()
		ret.reportText = TT.getDefaultText(recordNameNode.parent().type) + ' ' + recordNameNode.text
		def whereNode = recordNameNode.findDirectChild(TT.WHERE)
		JPNode orNode = null
		JPNode lastNode = null
		JPNode insertWhereBeforeNode = null
		if (whereNode)
			orNode = whereNode.findDirectChild(TT.OR)
		if (orNode)
			lastNode = whereNode.lastDescendant()

		if (orNode) {
			// We need to insert after the WHERE, and also a end parenthesis after the
			// last node in the WHERE's expression (because of the OR).
			ret.preprocessorProblems = ! matchList([recordNameNode, whereNode, lastNode])
		} else if (whereNode) {
			ret.preprocessorProblems = ! matchList([recordNameNode, whereNode])
		} else {
			insertWhereBeforeNode = getInsertWhereBeforeNode(recordNameNode)
			if (! insertWhereBeforeNode || ! insertWhereBeforeNode.text)
				ret.preprocessorProblems = true
			else
				ret.preprocessorProblems = ! matchList([recordNameNode, insertWhereBeforeNode])
		}

		if (ret.preprocessorProblems || RECORDNAME_DISABLED)
			return ret


		///// FROM HERE DOWN MUST BE ATOMIC (ALL OR NOTHING) /////
		currSourceFileNumChanges++

		String insertText = recordNameNode.text + ".entity EQ roadrunner.System:Entity"
		String indent = null
		if (whereNode) {
			String indent1 = getIndent(whereNode.line) // is line after, count by 0
			String indent2 = getIndent(recordNameNode.parent().line - 1)
			if (indent1.length() > indent2.length())
				indent = indent1
			else
				indent = indent2 + "    "
		}

		if (orNode) {
			// In case these are on the same line, do the last one first, to
			// avoid screwing up column offsets.
			appendToNode(lastNode, ")")
			appendToNode(whereNode, "AND (")
			
			appendIndentedNewLineToNode(whereNode, indent, insertText)
		} else if (whereNode) {
			appendToNode(whereNode, "AND")
			appendIndentedNewLineToNode(whereNode, indent, insertText)
		} else {
			String wherePhrase = "WHERE " + insertText
			char c = currSourceLines[insertWhereBeforeNode.line - 1]
					.charAt(insertWhereBeforeNode.column - 2)
			if (! Character.isWhitespace(c))
				wherePhrase = " " + wherePhrase
			c = currSourceLines[insertWhereBeforeNode.line - 1]
					.charAt(insertWhereBeforeNode.column - 1)
			if (c != ':' && c != '.')
				wherePhrase = wherePhrase + " "
			insertBeforeNode(insertWhereBeforeNode, wherePhrase)
		}

		return ret
	} // convertRecordName
	
	
	/** Get the whitespace indent for a line.
	 * Input line number, count from zero (node.line - 1).
	 */
	String getIndent(int line) {
		def ret = new StringBuilder()
		String text = currSourceLines[line]
		int i = 0
		int len = text.length()
		while (i < len && Character.isWhitespace(text.charAt(i))) {
			ret.append(text.charAt(i))
			++i
		}
		return ret.toString()
	}


	/** Get an appropriate node to insert a WHERE phrase in front of */
	JPNode getInsertWhereBeforeNode(JPNode recordNameNode) {
		def types = new HashSet<Integer>()
		types.addAll([TT.NOLOCK, TT.SHARELOCK, TT.EXCLUSIVELOCK, TT.NOPREFETCH, TT.NOERROR_KW])
		def child = recordNameNode.firstChild()
		while (child) {
			if (types.contains(child.type))
				return child
			child = child.nextSibling()
		}
		def next = recordNameNode.nextSibling()
		if (next)
			return next
		return recordNameNode.parent().nextSibling()
	}
	
	
	/** Get the text on the line after a node. */
	String getTextAfter(JPNode node) {
		String lineText = currSourceLines[node.line - 1]
		int column = node.column + node.text.length()
		return lineText.substring(column)
	}
	

	/** Insert text before a node's text. */
	void insertBeforeNode(JPNode node, String insertText) {
		int line = node.line - 1
		String oldLine = currSourceLines[line]
		int column = node.column - 1
		currSourceLines[line] =
			oldLine.substring(0, column) +
			insertText +
			oldLine.substring(column)
	}


	void loadIfNotCurrent(String filename) {
		// Find the (possibly relative) file name on the propath.
		def newfile = FileStuff.findFile(filename)
		if (newfile.equals(currSourceFile))
			return
		if (currSourceFileNumChanges > 0)
			writeCurrent()
		currSourceFileNumChanges = 0
		currSourceFile = newfile
		currSourceLines.clear()
		currSourceLines.addAll(newfile.readLines())
	}


	void logNoConvert(JPNode node, String message) {
		logger.linebreak()
		logger.logLine(node.filename + ' line ' + node.line)
		logger.logLine('    ' + message)
	}


	/** Test that a node's text can be matched to the source file text.
	 * Returns true if the input node is null.
	 */
	boolean match(JPNode node) {
		if (!node)
			return true
		// Proparse counts from 1, but we need line and column indexes counting from zero.
		def line = (String) currSourceLines[node.line - 1]
		int column = node.column - 1
		def text = node.text
		def end = column + text.length()
		if (end > line.length())
			return false
		return text.equals(line.substring(column, end))
	}


	/** Call match(node) for all nodes in a list.
	 * All nodes must be in the same file for the match to succeed.
	 * An empty list will return true.
	 */
	@SuppressWarnings("unchecked")
	boolean matchList(List nodes) {
		if (nodes.size() == 0)
			return true
		def firstNode = (JPNode) nodes.get(0)
		int fileIndex = firstNode.getFileIndex()
		for (JPNode node in nodes) {
			if (! match(node))
				return false
			if (node.getFileIndex() != fileIndex)
				return false
		}
		return true
	}


	/** Test if a CREATE node meets the criteria for conversion.
	 * We are looking for record phrases, from any statement that uses
	 * a record phrase. There is no "Record_phrase" in the tree, unfortunately.
	 */
	boolean testCreateNode(JPNode node) {
		if (node.firstChild().type != TT.RECORD_NAME)
			return false
		if (! support.testNodeIsAppSource(node))
			return false
		// Proparse uses an int flag to indicate if the
		// buffer's "store type" is for a WORK-TABLE, TEMP-TABLE, or db table.
		// We aren't interested in TEMP or WORK tables.
		def recordNameNode = (RecordNameNode) node.firstChild()
		if (recordNameNode.tableBuffer.table.storetype != IConstants.ST_DBTABLE)
			return false
		if (! recordNameNode.tableBuffer.table.database.name.equals("mroad"))
			return false
		return true
	}


	/** Test if an INSERT statement needs looking at.
	 * These are not auto-converted, but if any of them exist, they need
	 * to be logged.
	 */
	boolean testInsertNode(JPNode insertNode) {
		if (! insertNode.isStateHead())
			return false
		def recordNode = (RecordNameNode) insertNode.firstChild()
		if (recordNode.tableBuffer.table.storetype != IConstants.ST_DBTABLE)
			return false
		return true
	}


	/** Test if a RECORD_NAME node meets the criteria for conversion.
	 * We are looking for record phrases, from any statement that uses
	 * a record phrase. There is no "Record_phrase" in the tree, unfortunately.
	 */
	boolean testRecordNameNode(RecordNameNode node) {
		def isRecordPhrase = false
		switch (node.parent().type) {
			case TT.CANFIND:
			case TT.FIND:
			case TT.OPEN:
			case TT.PRESELECT:
			case TT.FOR:
				isRecordPhrase = true
		}
		// A "DO FOR ..." or "REPEAT FOR ..." is not a record phrase, and
		// does not need changing.
		if (node.parent().type == TT.FOR) {
			int t = node.parent().parent().type
			if (t == TT.DO || t == TT.REPEAT)
				return false
		}
		if (! (isRecordPhrase && support.testNodeIsAppSource(node)))
			return false
		// Proparse uses an int flag to indicate if the
		// buffer's "store type" is for a WORK-TABLE, TEMP-TABLE, or db table.
		// We aren't interested in TEMP or WORK tables.
		if (node.tableBuffer.table.storetype != IConstants.ST_DBTABLE)
			return false
		if (! node.tableBuffer.table.database.name.equals("mroad"))
			return false
		// We don't have to change a recordphrase with OF in it.
		if (node.findDirectChild(TT.OF))
			return false
		if (testRecordNameNodeIsJustWhereRecid(node))
			return false
		return true
	}


	boolean testRecordNameNodeIsJustWhereRecid(RecordNameNode recordNameNode) {
		//    RECORD_NAME
		//      WHERE
		//        EQ
		//          RECID
		//            LEFTPAREN
		//            RECORD_NAME
		//            RIGHTPAREN
		//          Field_ref
		def whereNode = recordNameNode.firstChild()
		if (whereNode == null || whereNode.type != TT.WHERE)
			return false
		if (whereNode.numberOfChildren != 1)
			return false
		def eqNode = whereNode.firstChild()
		if (eqNode.type != TT.EQ)
			return false
		if (eqNode.numberOfChildren != 2)
			return false
		def recidNode = eqNode.firstChild()
		if (recidNode.type != TT.RECID)
			return false
		if (recidNode.numberOfChildren != 3)
			return false
		def rnn2 = recidNode.findDirectChild(TT.RECORD_NAME)
		if (! rnn2)
			return false
		if (! rnn2.text.equalsIgnoreCase(recordNameNode.text))
			return false
		return true
	}


	/** Write the current (modified) source file to the output directory. */
	void writeCurrent() {
		totalChanges += currSourceFileNumChanges

		String outfileRelPath
		totalChangedFiles++
		outfileRelPath = currSourceFile.canonicalPath.substring(support.inputDir.canonicalPath.length())

		def outFile = new File(support.outputDir.path + File.separator + outfileRelPath)
		outFile.parentFile.mkdirs()
		// Empty the file, in case we're overwriting a previous run.
		outFile.write('')
		for (line in currSourceLines) {
			if (line!=null) {
				outFile << line
				outFile << LINESEP
			}
		}
	}


}
